==> lihil-0.2.24.tar.gz: lihil-0.2.24/tests/test_plugins/test_premier.py <==
import asyncio
import json
import time
from unittest.mock import AsyncMock, patch

import pytest
from premier.providers import AsyncInMemoryCache
from premier.throttler.errors import QuotaExceedsError

from lihil.local_client import LocalClient
from lihil.plugins.premier import PremierPlugin, Throttler


async def get_body_str(result):
    """Helper to get string body from result"""
    body_bytes = await result.body()
    return json.loads(body_bytes.decode())


async def test_throttling():
    async def hello():
        print("called the hello func")
        return "hello"

    lc = LocalClient()

    throttler = Throttler()

    plugin = PremierPlugin(throttler=throttler)

    ep = await lc.make_endpoint(hello, plugins=[plugin.fix_window(1, 1)])

    await lc(ep)

    with pytest.raises(QuotaExceedsError):
        for _ in range(2):
            await lc(ep)


async def test_fixed_window():
    """Test fixed window rate limiting"""

    async def api_call():
        return "success"

    lc = LocalClient()
    plugin = PremierPlugin()

    ep = await lc.make_endpoint(api_call, plugins=[plugin.fixed_window(2, 1)])

    # First two calls should succeed
    result1 = await lc(ep)
    result2 = await lc(ep)
    assert await get_body_str(result1) == "success"
    assert await get_body_str(result2) == "success"

    # Third call should be throttled
    with pytest.raises(QuotaExceedsError):
        await lc(ep)


async def test_cache_basic():
    """Test basic caching functionality"""
    call_count = 0

    async def expensive_operation():
        nonlocal call_count
        call_count += 1
        return f"computed_{call_count}"

    lc = LocalClient()
    plugin = PremierPlugin()

    ep = await lc.make_endpoint(expensive_operation, plugins=[plugin.cache(expire_s=1)])

    # First call should execute the function
    result1 = await lc(ep)
    assert await get_body_str(result1) == "computed_1"
    assert call_count == 1

    # Second call should return cached result
    result2 = await lc(ep)
    assert await get_body_str(result2) == "computed_1"  # Same result, from cache
    assert call_count == 1  # Function not called again


async def test_cache_with_ttl():
    """Test cache with TTL expiration"""
    call_count = 0
    mock_time = [1000.0]  # Use list to make it mutable

    def mock_timer():
        return mock_time[0]

    async def time_sensitive_operation():
        nonlocal call_count
        call_count += 1
        return f"result_{call_count}"

    lc = LocalClient()

    # Create cache provider with custom timer function for testing
    custom_cache = AsyncInMemoryCache(timer_func=mock_timer)
    plugin = PremierPlugin(cache_provider=custom_cache)

    ep = await lc.make_endpoint(
        time_sensitive_operation, plugins=[plugin.cache(expire_s=1)]
    )

    # First call
    result1 = await lc(ep)
    assert await get_body_str(result1) == "result_1"
    assert call_count == 1

    # Second call should use cache (same time)
    result2 = await lc(ep)
    assert await get_body_str(result2) == "result_1"
    assert call_count == 1

    # Simulate time passing beyond cache expiry
    mock_time[0] = 1002.0  # 2 seconds later (> 1s expiry)

    # Third call should execute function again due to cache expiry
    result3 = await lc(ep)
    assert await get_body_str(result3) == "result_2"
    assert call_count == 2


async def test_cache_with_custom_key():
    """Test cache with custom key generation"""
    call_count = 0

    async def user_operation(user_id: str):
        nonlocal call_count
        call_count += 1
        return f"user_{user_id}_data_{call_count}"

    lc = LocalClient()
    plugin = PremierPlugin()

    # Use custom key function
    ep = await lc.make_endpoint(
        user_operation,
        plugins=[plugin.cache(cache_key=lambda user_id: f"user:{user_id}")],
    )

    # Calls with same user_id should be cached
    result1 = await lc(ep, query_params={"user_id": "123"})
    result2 = await lc(ep, query_params={"user_id": "123"})
    assert await get_body_str(result1) == await get_body_str(result2)
    assert call_count == 1

    # Call with different user_id should execute function
    result3 = await lc(ep, query_params={"user_id": "456"})
    assert await get_body_str(result3) != await get_body_str(result1)
    assert call_count == 2


async def test_retry_basic():
    """Test basic retry functionality"""
    attempt_count = 0

    async def flaky_service():
        nonlocal attempt_count
        attempt_count += 1
        if attempt_count < 3:
            raise ConnectionError("Connection failed")
        return "success"

    lc = LocalClient()
    plugin = PremierPlugin()

    ep = await lc.make_endpoint(
        flaky_service, plugins=[plugin.retry(max_attempts=3, wait=0.001)]
    )

    result = await lc(ep)
    assert await get_body_str(result) == "success"
    assert attempt_count == 3


async def test_retry_with_exponential_backoff():
    """Test retry with exponential backoff"""
    attempt_count = 0

    async def unreliable_service():
        nonlocal attempt_count
        attempt_count += 1
        if attempt_count < 3:
            raise TimeoutError("Service timeout")
        return "recovered"

    lc = LocalClient()
    plugin = PremierPlugin()

    # Use minimal wait times to eliminate delays
    ep = await lc.make_endpoint(
        unreliable_service, plugins=[plugin.retry(max_attempts=3, wait=[0.001, 0.002])]
    )

    result = await lc(ep)
    assert await get_body_str(result) == "recovered"
    assert attempt_count == 3


async def test_retry_specific_exceptions():
    """Test retry with specific exception types"""
    attempt_count = 0

    async def service_with_different_errors():
        nonlocal attempt_count
        attempt_count += 1
        if attempt_count == 1:
            raise ConnectionError("Network error")  # Should retry
        elif attempt_count == 2:
            raise ValueError("Invalid data")  # Should not retry
        return "success"

    lc = LocalClient()
    plugin = PremierPlugin()

    ep = await lc.make_endpoint(
        service_with_different_errors,
        plugins=[plugin.retry(max_attempts=3, exceptions=(ConnectionError,), wait=0.001)],
    )

    # Should fail on ValueError without retrying
    with pytest.raises(ValueError):
        await lc(ep)

    assert attempt_count == 2  # One initial call + one retry


async def test_retry_with_on_fail_callback():
    """Test retry with failure callback"""
    attempt_count = 0
    failure_logs = []

    async def log_failure(*args, **kwargs):
        failure_logs.append(f"Failed attempt {attempt_count}")

    async def always_failing_service():
        nonlocal attempt_count
        attempt_count += 1
        raise RuntimeError("Always fails")

    lc = LocalClient()
    plugin = PremierPlugin()

    ep = await lc.make_endpoint(
        always_failing_service,
        plugins=[plugin.retry(max_attempts=3, wait=0.001, on_fail=log_failure)],
    )

    with pytest.raises(RuntimeError):
        await lc(ep)

    assert attempt_count == 3
    assert len(failure_logs) == 2  # Called on first 2 failures, not the final one


async def test_timeout():
    """Test timeout functionality"""

    async def slow_operation():
        # Mock slow operation that would timeout
        await asyncio.sleep(0.01)  # Very short delay for testing
        return "completed"

    lc = LocalClient()
    throttler = Throttler()
    plugin = PremierPlugin(throttler=throttler)

    # Mock the timer module's await_for function
    with patch('premier.timer.timer.await_for') as mock_await_for:
        mock_await_for.side_effect = asyncio.TimeoutError()

        ep = await lc.make_endpoint(
            slow_operation, plugins=[plugin.timeout(1)]  # 1 second timeout
        )

        with pytest.raises(asyncio.TimeoutError):
            await lc(ep)


async def test_timeout_with_logger():
    """Test timeout with logging"""
    logged_messages: list[str] = []

    class MockLogger:
        def exception(self, msg: str):
            logged_messages.append(msg)

        def info(self, msg):
            pass

    async def slow_operation():
        # Mock slow operation that would timeout
        await asyncio.sleep(0.01)  # Very short delay for testing
        return "completed"

    lc = LocalClient()
    throttler = Throttler()
    plugin = PremierPlugin(throttler=throttler)
    mock_logger = MockLogger()

    # Mock the timer module's await_for function
    with patch('premier.timer.timer.await_for') as mock_await_for:
        mock_await_for.side_effect = asyncio.TimeoutError()

        ep = await lc.make_endpoint(
            slow_operation, plugins=[plugin.timeout(1, logger=mock_logger)]
        )

        # Test that timeout works with logger parameter (even if logging doesn't work as expected)
        with pytest.raises((TimeoutError, asyncio.TimeoutError)):
            await lc(ep)

        # Note: The logger functionality may work differently in practice
        # This test verifies the timeout decorator accepts a logger parameter


async def test_custom_cache_provider():
    """Test plugin with custom cache provider"""
    custom_cache = AsyncInMemoryCache()
    plugin = PremierPlugin(cache_provider=custom_cache)

    call_count = 0

    async def cacheable_operation():
        nonlocal call_count
        call_count += 1
        return f"result_{call_count}"

    lc = LocalClient()
    ep = await lc.make_endpoint(cacheable_operation, plugins=[plugin.cache()])

    # First call should execute
    result1 = await lc(ep)
    assert await get_body_str(result1) == "result_1"
    assert call_count == 1

    # Second call should use cache
    result2 = await lc(ep)
    assert await get_body_str(result2) == "result_1"
    assert call_count == 1


async def test_combined_features():
    """Test combining multiple plugin features"""
    call_count = 0
    attempt_count = 0

    async def complex_service(data: str):
        nonlocal call_count, attempt_count
        attempt_count += 1

        # Fail on first attempt to test retry
        if call_count == 0 and attempt_count == 1:
            raise ConnectionError("First attempt fails")

        call_count += 1
        # Remove the sleep to eliminate delay
        return f"processed_{data}_{call_count}"

    lc = LocalClient()
    plugin = PremierPlugin()

    # Combine timeout, retry, cache, and throttling
    ep = await lc.make_endpoint(
        complex_service,
        plugins=[
            plugin.timeout(2),  # 2 second timeout
            plugin.retry(max_attempts=2, wait=0.001),  # Retry once with minimal delay
            plugin.cache(expire_s=1),  # 1 second cache
            plugin.fixed_window(5, 1),  # 5 requests per second
        ],
    )

    # First call should succeed after retry
    result1 = await lc(ep, query_params={"data": "test"})
    assert await get_body_str(result1) == "processed_test_1"
    assert attempt_count == 2  # Failed once, then succeeded
    assert call_count == 1

    # Second call should use cache (same function not called again)
    result2 = await lc(ep, query_params={"data": "test"})
    assert await get_body_str(result2) == "processed_test_1"  # Same cached result
    assert call_count == 1  # Function not called again


async def test_backward_compatibility():
    """Test that fix_window alias still works"""

    async def hello():
        return "hello"

    lc = LocalClient()
    throttler = Throttler()
    plugin = PremierPlugin(throttler=throttler)

    # Test the old fix_window method still works
    ep = await lc.make_endpoint(hello, plugins=[plugin.fix_window(1, 1)])

    result = await lc(ep)
    assert await get_body_str(result) == "hello"

    # Should be throttled on second call
    with pytest.raises(QuotaExceedsError):
        await lc(ep)

==> lihil-0.2.33.tar.gz: lihil-0.2.33/tests/test_plugins/test_premier.py <==
import asyncio
from unittest.mock import patch

import pytest

pytestmark = pytest.mark.requires_premier

try:
    from premier.providers import AsyncInMemoryCache
    from premier.throttler.errors import QuotaExceedsError
    from lihil.plugins.premier import PremierPlugin, Throttler
except ModuleNotFoundError as exc:
    pytest.skip(f"premier dependencies unavailable: {exc}", allow_module_level=True)

from lihil.local_client import LocalClient


async def test_throttling():
    async def hello():
        print("called the hello func")
        return "hello"

    lc = LocalClient()

    throttler = Throttler()

    plugin = PremierPlugin(throttler=throttler)

    ep = await lc.make_endpoint(hello, plugins=[plugin.fix_window(1, 1)])

    await lc(ep)

    with pytest.raises(QuotaExceedsError):
        for _ in range(2):
            await lc(ep)


async def test_fixed_window():
    """Test fixed window rate limiting"""
    async def api_call():
        return "success"

    lc = LocalClient()
    plugin = PremierPlugin()

    ep = await lc.make_endpoint(api_call, plugins=[plugin.fixed_window(2, 1)])

    # First two calls should succeed
    result1 = await lc(ep)
    result2 = await lc(ep)
    assert await result1.json() == "success"
    assert await result2.json() == "success"

    # Third call should be throttled
    with pytest.raises(QuotaExceedsError):
        await lc(ep)


async def test_cache_basic():
    """Test basic caching functionality"""
    call_count = 0

    async def expensive_operation():
        nonlocal call_count
        call_count += 1
        return f"computed_{call_count}"

    lc = LocalClient()
    plugin = PremierPlugin()

    ep = await lc.make_endpoint(expensive_operation, plugins=[plugin.cache(expire_s=1)])

    # First call should execute the function
    result1 = await lc(ep)
    assert await result1.json() == "computed_1"
    assert call_count == 1

    # Second call should return cached result
    result2 = await lc(ep)
    assert await result2.json() == "computed_1"  # Same result, from cache
    assert call_count == 1  # Function not called again


async def test_cache_with_ttl():
    """Test cache with TTL expiration"""
    call_count = 0
    mock_time = [1000.0]  # Use list to make it mutable

    def mock_timer():
        return mock_time[0]

    async def time_sensitive_operation():
        nonlocal call_count
        call_count += 1
        return f"result_{call_count}"

    lc = LocalClient()

    # Create cache provider with custom timer function for testing
    custom_cache = AsyncInMemoryCache(timer_func=mock_timer)
    plugin = PremierPlugin(cache_provider=custom_cache)

    ep = await lc.make_endpoint(
        time_sensitive_operation, plugins=[plugin.cache(expire_s=1)]
    )

    # First call
    result1 = await lc(ep)
    assert await result1.json() == "result_1"
    assert call_count == 1

    # Second call should use cache (same time)
    result2 = await lc(ep)
    assert await result2.json() == "result_1"
    assert call_count == 1

    # Simulate time passing beyond cache expiry
    mock_time[0] = 1002.0  # 2 seconds later (> 1s expiry)

    # Third call should execute function again due to cache expiry
    result3 = await lc(ep)
    assert await result3.json() == "result_2"
    assert call_count == 2


async def test_cache_with_custom_key():
    """Test cache with custom key generation"""
    call_count = 0

    async def user_operation(user_id: str):
        nonlocal call_count
        call_count += 1
        return f"user_{user_id}_data_{call_count}"

    lc = LocalClient()
    plugin = PremierPlugin()

    # Use custom key function
    ep = await lc.make_endpoint(
        user_operation,
        plugins=[plugin.cache(cache_key=lambda user_id: f"user:{user_id}")],
    )

    # Calls with same user_id should be cached
    result1 = await lc(ep, query_params={"user_id": "123"})
    result2 = await lc(ep, query_params={"user_id": "123"})
    assert await result1.json() == await result2.json()
    assert call_count == 1

    # Call with different user_id should execute function
    result3 = await lc(ep, query_params={"user_id": "456"})
    assert await result3.json() != await result1.json()
    assert call_count == 2


async def test_retry_basic():
    """Test basic retry functionality"""
    attempt_count = 0

    async def flaky_service():
        nonlocal attempt_count
        attempt_count += 1
        if attempt_count < 3:
            raise ConnectionError("Connection failed")
        return "success"

    lc = LocalClient()
    plugin = PremierPlugin()

    ep = await lc.make_endpoint(
        flaky_service, plugins=[plugin.retry(max_attempts=3, wait=0.001)]
    )

    result = await lc(ep)
    assert await result.json() == "success"
    assert attempt_count == 3


async def test_retry_with_exponential_backoff():
    """Test retry with exponential backoff"""
    attempt_count = 0

    async def unreliable_service():
        nonlocal attempt_count
        attempt_count += 1
        if attempt_count < 3:
            raise TimeoutError("Service timeout")
        return "recovered"

    lc = LocalClient()
    plugin = PremierPlugin()

    # Use minimal wait times to eliminate delays
    ep = await lc.make_endpoint(
        unreliable_service, plugins=[plugin.retry(max_attempts=3, wait=[0.001, 0.002])]
    )

    result = await lc(ep)
    assert await result.json() == "recovered"
    assert attempt_count == 3


async def test_retry_specific_exceptions():
    """Test retry with specific exception types"""
    attempt_count = 0

    async def service_with_different_errors():
        nonlocal attempt_count
        attempt_count += 1
        if attempt_count == 1:
            raise ConnectionError("Network error")  # Should retry
        elif attempt_count == 2:
            raise ValueError("Invalid data")  # Should not retry
        return "success"

    lc = LocalClient()
    plugin = PremierPlugin()

    ep = await lc.make_endpoint(
        service_with_different_errors,
        plugins=[
            plugin.retry(max_attempts=3, exceptions=(ConnectionError,), wait=0.001)
        ],
    )

    # Should fail on ValueError without retrying
    with pytest.raises(ValueError):
        await lc(ep)

    assert attempt_count == 2  # One initial call + one retry


async def test_retry_with_on_fail_callback():
    """Test retry with failure callback"""
    attempt_count = 0
    failure_logs = []

    async def log_failure(*args, **kwargs):
        failure_logs.append(f"Failed attempt {attempt_count}")

    async def always_failing_service():
        nonlocal attempt_count
        attempt_count += 1
        raise RuntimeError("Always fails")

    lc = LocalClient()
    plugin = PremierPlugin()

    ep = await lc.make_endpoint(
        always_failing_service,
        plugins=[plugin.retry(max_attempts=3, wait=0.001, on_fail=log_failure)],
    )

    with pytest.raises(RuntimeError):
        await lc(ep)

    assert attempt_count == 3
    assert len(failure_logs) == 2  # Called on first 2 failures, not the final one


async def test_timeout():
    """Test timeout functionality"""
    async def slow_operation():
        # Mock slow operation that would timeout
        await asyncio.sleep(0.01)  # Very short delay for testing
        return "completed"

    lc = LocalClient()
    throttler = Throttler()
    plugin = PremierPlugin(throttler=throttler)

    # Mock the timer module's await_for function
    with patch("premier.timer.timer.await_for") as mock_await_for:
        mock_await_for.side_effect = asyncio.TimeoutError()

        ep = await lc.make_endpoint(
            slow_operation, plugins=[plugin.timeout(1)]  # 1 second timeout
        )

        with pytest.raises(asyncio.TimeoutError):
            await lc(ep)


async def test_timeout_with_logger():
    """Test timeout with logging"""
    logged_messages: list[str] = []

    class MockLogger:
        def exception(self, msg: str):
            logged_messages.append(msg)

        def info(self, msg):
            pass

    async def slow_operation():
        # Mock slow operation that would timeout
        await asyncio.sleep(0.01)  # Very short delay for testing
        return "completed"

    lc = LocalClient()
    throttler = Throttler()
    plugin = PremierPlugin(throttler=throttler)
    mock_logger = MockLogger()

    # Mock the timer module's await_for function
    with patch("premier.timer.timer.await_for") as mock_await_for:
        mock_await_for.side_effect = asyncio.TimeoutError()

        ep = await lc.make_endpoint(
            slow_operation, plugins=[plugin.timeout(1, logger=mock_logger)]
        )

        # Test that timeout works with logger parameter (even if logging doesn't work as expected)
        with pytest.raises((TimeoutError, asyncio.TimeoutError)):
            await lc(ep)

        # Note: The logger functionality may work differently in practice
        # This test verifies the timeout decorator accepts a logger parameter


async def test_custom_cache_provider():
    """Test plugin with custom cache provider"""
    custom_cache = AsyncInMemoryCache()
    plugin = PremierPlugin(cache_provider=custom_cache)

    call_count = 0

    async def cacheable_operation():
        nonlocal call_count
        call_count += 1
        return f"result_{call_count}"

    lc = LocalClient()
    ep = await lc.make_endpoint(cacheable_operation, plugins=[plugin.cache()])

    # First call should execute
    result1 = await lc(ep)
    assert await result1.json() == "result_1"
    assert call_count == 1

    # Second call should use cache
    result2 = await lc(ep)
    assert await result2.json() == "result_1"
    assert call_count == 1


async def test_combined_features():
    """Test combining multiple plugin features"""
    call_count = 0
    attempt_count = 0

    async def complex_service(data: str):
        nonlocal call_count, attempt_count
        attempt_count += 1

        # Fail on first attempt to test retry
        if call_count == 0 and attempt_count == 1:
            raise ConnectionError("First attempt fails")

        call_count += 1
        # Remove the sleep to eliminate delay
        return f"processed_{data}_{call_count}"

    lc = LocalClient()
    plugin = PremierPlugin()

    # Combine timeout, retry, cache, and throttling
    ep = await lc.make_endpoint(
        complex_service,
        plugins=[
            plugin.timeout(2),  # 2 second timeout
            plugin.retry(max_attempts=2, wait=0.001),  # Retry once with minimal delay
            plugin.cache(expire_s=1),  # 1 second cache
            plugin.fixed_window(5, 1),  # 5 requests per second
        ],
    )

    # First call should succeed after retry
    result1 = await lc(ep, query_params={"data": "test"})
    assert await result1.json() == "processed_test_1"
    assert attempt_count == 2  # Failed once, then succeeded
    assert call_count == 1

    # Second call should use cache (same function not called again)
    result2 = await lc(ep, query_params={"data": "test"})
    assert await result2.json() == "processed_test_1"  # Same cached result
    assert call_count == 1  # Function not called again


async def test_backward_compatibility():
    """Test that fix_window alias still works"""
    async def hello():
        return "hello"

    lc = LocalClient()
    throttler = Throttler()
    plugin = PremierPlugin(throttler=throttler)

    # Test the old fix_window method still works
    ep = await lc.make_endpoint(hello, plugins=[plugin.fix_window(1, 1)])

    result = await lc(ep)
    assert await result.json() == "hello"

    # Should be throttled on second call
    with pytest.raises(QuotaExceedsError):
        await lc(ep)
